/* $Id: forkpty.c,v 1.1 1998/09/06 21:04:15 dps Exp $ */
/* forkpty */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_sTAT_H */
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif /* HAVE_SYS_IOCTL_H */
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif /* HAVE_TERMIOS_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif /* HAVE_SIGNAL_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */

#include "libpt.h"
#undef openpty
#undef forkpty

int __libpt__forkpty(int *__masterfd, const char *name,
	    struct termios *__termp, struct winsize *__winp)
{
    int slave, pid, fd, p[2], succ;
    __sighandler_t hup_action;
    char c;

    if (pipe(p)==-1)				/* Used to talk to slave */
	return -1;

    if (__libpt__openpty(__masterfd, &slave, name, __termp, __winp)==-1)
	return -1;
    switch(pid=fork())
    {
    case -1:
	close(*__masterfd);
	close(slave);
	return -1;

    case 0:
	/* child */
	close(p[0]);				/* Close reading end */
	close(*__masterfd);			/* Close pty master */
	if (setsid()==-1)
	{
	    c=errno;				/* Save errno */
	    write(p[1], &c, 1);			/* Tell parent */
	    _exit(1);				/* Exit */
	}
	/* Drop current tty */
	if ((fd=open("/dev/tty", O_RDWR))!=-1)
	{
	    hup_action=signal(SIGHUP, SIG_IGN);	/* Ignore HUP */
	    succ=ioctl(fd, TIOCNOTTY, 0);	/* Can generate HUP */
	    signal(SIGHUP, hup_action);		/* Restore HUP handling */
	    close(fd);				/* Close /dev/tty */
	}
	else
	    succ=0;
	if (succ==-1)
	{
	    c=errno;				/* Save errno */
	    write(p[1], &c, 1);			/* Tell parent */
	    _exit(1);				/* Exit */
	}
	/* Set tty */
	if (ioctl(slave, TIOCSCTTY, 0)==-1)	/* Set to slave */
	{
	    c=errno;				/* Save errno */
	    write(p[1], &c, 1);			/* Tell parent */
	    _exit(1);				/* Exit */
	}
	dup2(slave, 0);				/* stdin is slave */
	dup2(slave, 1);				/* stdout is slave */
	dup2(slave, 2);				/* stderr is skave */
	if (slave>2)
	    close(slave);
	write(p[1], "\0", 1);			/* Success */
	close(p[1]);				/* Close pipe */
	return 0;				/* Child */
	/* Not reached */

    default:
	/* parent */
	close(p[1]);				/* Close writing end */
	read(p[0], &c, 1);			/* REad code from slave */
	close(p[0]);				/* Close reading end of pipe */
	if (c!=0)
	{
	    waitpid(pid, &succ, 0);		/* Wait for child */
	    close(*__masterfd);			/* Close pty master */
	    errno=c;				/* Set errno */
	    return -1;				/* Fail */
	}
	break;
    }
    return pid;					/* Parent */
}

