/* $Id: openpty.c,v 1.1 1998/09/06 21:02:42 dps Exp $ */
/* openpty siimulation for ptmx */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_sTAT_H */
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif /* HAVE_SYS_IOCTL_H */
#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif /* HAVE_TERMIOS_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_STTINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */

#include "libpt.h"
#undef openpty
#undef forkpty

#define MAX_PTSNUM 9999

int __libpt__openpty(int *__masterfd, int *__slavefd, const char *name,
	    struct termios *__termp, struct winsize *__winp)
{
    const int unlock=0;		/* Constants for ioctl */
    int ptynum;			/* slave pty number */
    char slave_name[15];	/* /dev/pts/#### is 14 chars */


    if ((*__masterfd=open("/dev/ptmx", O_RDWR))==-1) /* Open master */
	return -1;

    /* Get name of slave */
    if (ioctl(*__masterfd, TIOCGPTN, &ptynum)==-1)
    {
	close(*__masterfd);
	return -1;
    }
    if (ptynum<0 || ptynum>MAX_PTSNUM)
    {
	close(*__masterfd);
	return -1;
    }
    sprintf(name, "/dev/pts/%d", ptynum);

    /* unlock and open slave */
    if (ioctl(*__masterfd, TIOCSPTLCK, &unlock)==-1)
    {
	close(*__masterfd);
	return -1;
    }
    if ((*__slavefd=open(slave_name, O_RDWR))==-1)
    {
	close(*__masterfd);
	return -1;
    }

    /* If specified set terminal characteristics */
    if (__termp!=NULL && tcsetattr(*__slavefd, TCSANOW, __termp)==-1)
    {
	close(*__masterfd);
	close(*__slavefd);
	return -1;
    }
    if (__winp!=NULL && ioctl(*__slavefd, TIOCSWINSZ, __winp)==-1)
    {
	close(*__masterfd);
	close(*__slavefd);
	return -1;
    }
    if (name!=NULL)
    {
	strcpy(name, slave_name);
    }
    return 0;
}
